//
//    Skeleton Frame processing routine Template
//
#include "DatTemplate.hh"
#include "FilterDesign.hh"

#ifndef __CINT__

//-->  The next three lines are needed if you are going to generate triggers.
//     The text in PIDTITLE should describe the monitor function.
#define PIDCVSHDR "$Header: https://redoubt.ligo-wa.caltech.edu/svn/gds/trunk/DMT/Templates/DatTemplate.cc 6832 2013-06-18 23:20:26Z john.zweizig@LIGO.ORG $"
#define PIDTITLE  "Data monitor template"
#include "ProcIdent.hh"

#include "Dacc.hh"
#include <iostream>
#include <string>
#include <cstdlib>
#include <unistd.h>
#include <pwd.h>

using namespace std;

//======================================  Generate the main routine.
EXECDAT(DatTemplate)
#endif               // !def(__CINT__)

//======================================  Skeleton object constructor.
DatTemplate::DatTemplate(int argc, const char *argv[]) 
  : DatEnv(argc, argv), maxFrame(999999), mStep(2.0), mHistory(43200)
{
    //----------------------------------  Look for arguments
    bool dvIndex = false;

    bool syntax  = false;
    for (int i=1 ; i<argc ; i++) {
        string argi = argv[i];
        if (argi == "-channel") {
	    mChannel = argv[++i];
	} else if (argi == "-dvindex") {
	    dvIndex = true;
	} else if (argi == "-filter") {
	    mFilter = argv[++i];
	} else if (argi == "-frame") {
	    maxFrame = strtol(argv[++i], 0, 0);
	} else if (argi == "-stride") {
	    mStep = strtod(argv[++i], 0);
	} else if (argi == "-settle") {
	    mSettle = strtod(argv[++i], 0);
 	} else if (isDatEnvArg(argv[i])) {
	    i++;
	} else {
	    syntax = true;
	    cerr << "Unrecognized argument: " << argi << endl;
	}
    }

    //----------------------------------  Bail out if command line not correct
    if (syntax) {
        cerr << "Command syntax:" << endl;
	cerr << argv[0] 
	     << " [-channel <nMax>] [-frames <nMax>] [-stride <time>] \\" 
	     << endl
	     << "      [-filter <filter>] [-settle <s-time>] [-dvIndex] \\"
	     << endl
	     << "      [<DatEnv-args>]" 
	     << endl;
	finish();
	return;
    }

    //----------------------------------  Make sure there's a valid channel
    if (mChannel.empty()) mChannel = "H0:PEM-LVEA_SEISX";

    //----------------------------------  set the data accessor
    getDacc().setStride(mStep);       // in before calling ProcessData
    getDacc().addChannel(mChannel.c_str());

    //----------------------------------  Specify DMT Viewer channel names
    //--- replaced cuserid() with getpwuid - K. Thorne June, 2011
    struct passwd *pws;
    pws = getpwuid(geteuid());
    setServerName(pws->pw_name);
    
    mSigmaName = mChannel + "_sigma";
    serveData(mSigmaName.c_str(), &mHistory, 0);

    //----------------------------------  Set up the trender
    mTrend.setName("DatTemplate");    //  Set trend frame name
    mTrend.setFrameCount(1);          //  Set frames per file
    mTrend.setType(Trend::kMinute);   //  Set trend type (minute trends)
    mTrend.setAutoUpdate(false);      //  Avoid automatic screwup.
    mTrend.addChannel(mSigmaName.c_str());

    //----------------------------------  Optional dataviewer index
    if (dvIndex) {
        string filename;
	const char* dmtout = getenv("DMTOUTPUT");
	if (dmtout) {
	    filename = dmtout;
	    filename += "/";
	}
	filename += "channel.cfg";
	mTrend.writeIndex(filename.c_str());
    }
}

//======================================  Skeleton object destructor.
DatTemplate::~DatTemplate() 
{
    cout << "DatTemplate is finished" << endl;
}

//======================================  Frame processing function.
void
DatTemplate::ProcessData(void) {
    //----------------------------------  Get pointers to the current data.
    const TSeries* ts = getDacc().refData(mChannel.c_str());
    if (!ts || ts->isEmpty()) {
        cout << "Channel: " << mChannel << " was not found." << endl;
        return;
    }

    //----------------------------------  Construct the filter if it isn't 
    //                                    already there
    if (mPipe.null() && !mFilter.empty()) {
        double samplerate(1./double(ts->getTStep()));
	FilterDesign fd(mFilter.c_str(), samplerate);
	mPipe.set(fd.release());
    }

    //----------------------------------  Filter the data
    TSeries tf(mPipe(*ts));

    //----------------------------------  Calculate channel sigma.
    float Sig = sqrt( tf*tf/tf.getNSample() - pow(tf.getAverage(), 2) );

    //----------------------------------  Trend the data points
    Time t0 = tf.getStartTime();
    mTrend.trendData(mSigmaName.c_str(), t0, Sig);
    mTrend.Update();

    //----------------------------------  set up data for display.
    mHistory.fixedAppend(t0, tf.getInterval(), &Sig);
}


//======================================  Handle Messages
void
DatTemplate::Attention(void) {
    MonServer::Attention();
}
