/* -*- mode: c++; c-basic-offset: 3; -*- */
//======================================  W-pipe headers
#include "wplot.hh"
#include "wcondition.hh"
#include "weventlist.hh" 
#include "wfigure.hh"
#include "wframecache.hh"
#include "wmeasure.hh"
#include "wreaddata.hh"
#include "wresample.hh"
#include "wtile.hh"
#include "wtransform.hh"
#include "matlab_fcs.hh"

//======================================  DMT headers
#include "TSeries.hh"
#include "fSeries/DFT.hh"
#include "Interval.hh"
#include "frame_name.hh"

//======================================  C++ and system headers.
#include <fstream>
#include <sstream>
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <iomanip>
#include <stdexcept>
#include <algorithm>

using namespace std;
using namespace wpipe;

//======================================  Syntax output function
static void
wplot_syntax(void) {
   cerr << "The dmt_wplot command syntax is:" << endl;
}


//======================================  Plot names
const std::string evnt_raw   = "eventgram_raw";
const std::string evnt_white = "eventgram_whitened";
const std::string evnt_auto  = "eventgram_autoscaled";
const std::string spec_raw   = "spectrogram_raw";
const std::string spec_white = "spectrogram_whitened";
const std::string spec_auto  = "spectrogram_autoscaled";
const std::string tser_raw   = "timeseries_raw";
const std::string tser_white = "timeseries_whitened";
const std::string tser_hipas = "timeseries_highpassed";


//======================================  Main function.
int 
main(int argc, const char* argv[]) {
   plot_chan par;
   par.reset();

   try {
      scan_config(argc, argv, par);
   } catch (std::exception& e) {
      cerr << "Error reading configuration: " << e.what() << endl;
      return 1;
   }

   cout << "wplot parameters: " <<endl;
   par.display(cout);

   try {
      wplot(par);
   }
   catch (exception& e) {
      cerr << "caught exception: " << e.what() << endl;
   }
}

//======================================  plot function.
void
wpipe::wplot(const plot_chan& pchan) {

   /////////////////////////////////////////////////////////////////////////////
   //                        process command line arguments                   //
   /////////////////////////////////////////////////////////////////////////////
   string channelNameString = pchan.channelName;
   string frameType =         pchan.frameType;
   int    debugLevel =        pchan.debugLevel;

   // convert string event time and debug level to numbers
   ostringstream outstr;
   outstr << fixed << setprecision(3) << pchan.eventTime.totalS();
   string eventTimeString = outstr.str();

   // report status
   if (debugLevel >= 1) {
      time_t clock=0;
      cout << "OmegaScan of event at " << eventTimeString << endl;
      cout << "created by " << getenv("USER") << " on " << datestr(clock, 29)
	   <<" at " << datestr(clock, 13) << endl;
   }

   /////////////////////////////////////////////////////////////////////////////
   //                           load frame cache file                         //
   /////////////////////////////////////////////////////////////////////////////

   // report status
   if (debugLevel >= 1 && !pchan.frameCacheFile.empty()) {
      cout << "Building frame cache from: " << pchan.frameCacheFile << endl;
   }

   // load frame file cache
   wframecache frameCache(pchan.frameCacheFile);
   if (frameCache.empty() && (frameType == "NDS2" || frameType.empty())) {
      frameType = "NDS2";
      string nds_server = pchan.ndsServer;
      if (nds_server.empty()) nds_server = "nds.ligo.caltech.edu:31200";
      frameCache.add_group("H-NDS2", 0, 1999999999, 0, nds_server);
      frameCache.add_group("G-NDS2", 0, 1999999999, 0, nds_server);
      frameCache.add_group("L-NDS2", 0, 1999999999, 0, nds_server);
      frameCache.add_group("V-NDS2", 0, 1999999999, 0, nds_server);
   }

   if (debugLevel > 2) frameCache.display();

   /////////////////////////////////////////////////////////////////////////////
   //                         create output directory                         //
   /////////////////////////////////////////////////////////////////////////////

   // if (outputDirectory not specified, make one based on center time
   frame_name outFile("plots/%g", pchan.plotFile, 0, "");
   if (!pchan.outDir.empty()) outFile.set_directory(pchan.outDir);

   // report status
   string outputDirectory = outFile.dir_name(pchan.eventTime);
   if (debugLevel >= 1) {
      cout << "creating event directory" << endl;
      cout << "  outputDirectory:         " << outputDirectory << endl;
   }
   outFile.make_dir(outputDirectory, 2);

   ////////////////////////////////////////////////////////////////////////////
   //                       loop over configuration channels                 //
   ////////////////////////////////////////////////////////////////////////////
  
   resampler resample;
   resample.reset();
   dble_vect emptyRange;

   if (debugLevel > 1) pchan.display(cout);
  
   /////////////////////////////////////////////////////////////////////////
   //               identify statistically significant channels           //
   /////////////////////////////////////////////////////////////////////////
   double    sampleFrequency =      pchan.sampleFrequency;
   double    timeRange =            pchan.searchTimeRange;
   dble_vect frequencyRange =       pchan.searchFrequencyRange;
   dble_vect qRange =               pchan.searchQRange;
   double    maximumEnergyLoss =    pchan.searchMaximumEnergyLoss;
   double    searchWindowDuration = pchan.searchWindowDuration;

   // standardize channel names
   channelNameString = strrep(channelNameString, ";", ":");

   // Plot parameters
   dble_vect plotTimeRanges            = pchan.plotTimeRanges;
   dble_vect plotFrequencyRange        = pchan.plotFrequencyRange;
   dble_vect plotNormalizedEnergyRange = pchan.plotNormalizedEnergyRange;

   // find closest sample time to event time
   Time centerTime = Time(pchan.eventTime.getS());
   centerTime += round(double(pchan.eventTime-centerTime)*sampleFrequency)
               / sampleFrequency;

   // determine segment start and stop times
   Time startTime = centerTime - Interval(timeRange / 2.);
   if (startTime-Time(startTime.getS()) < Interval(0.5)) {
      startTime = Time(startTime.getS());
   } else {
      startTime = Time(startTime.getS()+1);
   }
   Time stopTime = startTime + timeRange;

   // generate search tiling
   wlog(debugLevel, 2, "  tiling for search");
   double highPassCutoff = -1;
   double lowPassCutoff = -1;
   double whiteningDuration = -1;

   wtile tiling(timeRange, qRange, frequencyRange, sampleFrequency, 
		maximumEnergyLoss, highPassCutoff, lowPassCutoff, 
		whiteningDuration, pchan.transientFactor);


   //  The threshold is set to yield the specified false event rate when 
   //  applied to all available frequencies and Qs of a white noise 
   //  signal. It is not modified to account for restricted f, Q ranges. 
   //  It is also only a rough estimate, and the result false event rate
   //  may vary significantly depending on the quality of the data.
   //
   double whiteNoiseFalseRate = pchan.whiteNoiseFalseRate;
   double eventThreshold = pchan.eventThreshold;
   if (eventThreshold == 0) {
      eventThreshold = tiling.threshold_from_rate(whiteNoiseFalseRate);
   }

   //========================================================================
   //  This has to be fixed for reasonable data I/O
   // read data from frame file
   wlog(debugLevel, 2, "  reading data");
   double timeShifts = 0;
   tser_vect rawData;
   wreaddata(frameCache, channelNameString, frameType, 
	     startTime, stopTime, timeShifts, debugLevel,
	     rawData);

   // check for read error
   if (rawData.empty()) {
      wlog(debugLevel, 1, "  ERROR: cannot load frame data");
      return;
   }

   // test multiple channels for consistency
   if (rawData.size() > 1) {
      wlog(debugLevel, 2, "  subtracting data");
      rawData[0] -= rawData[1];
      rawData.erase(rawData.begin()+1, rawData.end());
   }

   //--------------------------------  Check for all zeros
   if (rawData[0].getMinimum() == rawData[0].getMaximum()) {
      wlog(debugLevel, 1, "  WARNING: channel contains no information");
      return;
   }

   //--------------------------------  Resample data
   wlog(debugLevel, 2, "  resampling data");
   tser_vect rsamp;
   rsamp = resample.wresample(rawData, sampleFrequency);

   // high pass filter and whiten data
   wlog(debugLevel, 2, "  conditioning data");

   wcondition conditioner(rsamp, tiling);
   containers::DFT whitenedData = conditioner.whitenedDFT();

   // q transform whitened data
   wlog(debugLevel, 2, "  transforming whitened data");
   wtransform whitenedTransform(whitenedData, tiling, pchan.outlierFactor, 
				channelNameString);
   //whitenedTransform.dump(cout);

   // identify most significant whitened transform tile
   wlog(debugLevel, 2, "  measuring peak significance");
   Time thresholdReferenceTime = centerTime;
   dble_vect thresholdTimeRange(2);
   thresholdTimeRange[0] = -0.5*searchWindowDuration;
   thresholdTimeRange[1] =  0.5*searchWindowDuration;
   dble_vect thresholdQRange;
   dble_vect thresholdFrequencyRange;
   wmeasure whitenedProperties(whitenedTransform, tiling, startTime, 
			       thresholdReferenceTime, thresholdTimeRange, 
			       thresholdFrequencyRange, thresholdQRange, 
			       debugLevel);
   //whitenedProperties.dump(cout);

   //--------------------------------  If channel not significant
   if (whitenedProperties[0].peak.NormalizedEnergy < pchan.plotThreshold) {
      cerr << "Channel: " << channelNameString 
	   << " failed event thresold test, peak = " 
	   << whitenedProperties[0].peak.NormalizedEnergy
	   << " threshold = " << eventThreshold << endl;
      return;
   } // end test of channel significance

   /////////////////////////////////////////////////////////////////////////
   //               analyze statistically significant channels            //
   /////////////////////////////////////////////////////////////////////////
 
   //--------------------------------  Q transform raw data
   wlog(debugLevel, 2, "  transforming raw data");
   wtransform rawTransform(conditioner.rawDFT(), tiling, pchan.outlierFactor, 
			   channelNameString);
   //rawTransform.dump(cout);

   weventstack rawEvents, whitenedEvents;

   // identify significant whitened q transform tiles
   double maxPlotTRange = *(max_element(plotTimeRanges.begin(), 
					plotTimeRanges.end()));
   thresholdTimeRange[0] = -0.5*(maxPlotTRange 
				 + tiling.end_plane().row(0).duration 
				 * pchan.plotDurationInflation);
   thresholdTimeRange[1] = -thresholdTimeRange[0];
   thresholdFrequencyRange = plotFrequencyRange;
   thresholdQRange = emptyRange;

   //----------------  Identify significant tiles for whitened eventgram
   const_strv_iter begPlot = pchan.plotType.begin();
   const_strv_iter endPlot = pchan.plotType.end();

   if (find(begPlot, endPlot, evnt_white) != endPlot ||
       find(begPlot, endPlot, evnt_auto)  != endPlot) {

      wlog(debugLevel, 2, "  thresholding whitened transform");
      weventstack whitenedSignificants;
      whitenedSignificants.wthreshold(
			whitenedTransform, tiling, startTime, eventThreshold, 
			thresholdReferenceTime, thresholdTimeRange, 
			thresholdFrequencyRange, thresholdQRange, 
			pchan.maximumSignificants, "", 0, 0, 0, debugLevel);
      //whitenedSignificants.dump(cout);

      // select non-overlapping whitened significant tiles
      wlog(debugLevel, 2, "  selecting whitened events");
      whitenedEvents.wselect(whitenedSignificants, pchan.plotDurationInflation, 
			  pchan.plotBandwidthInflation,  pchan.maximumMosaics, 
			  debugLevel);
   }

   //---------------- Identify significant raw q transform tiles
   if (find(begPlot, endPlot, evnt_raw) != endPlot) {
      wlog(debugLevel, 2, "  thresholding raw transform");
      weventstack rawSignificants;
      rawSignificants.wthreshold(rawTransform, tiling, startTime, 
				 eventThreshold, thresholdReferenceTime,
				 thresholdTimeRange, thresholdFrequencyRange,
				 thresholdQRange, pchan.maximumSignificants, 
				 "", 0, 0, 0, debugLevel);
      //rawSignificants.dump(cout);

      // select non-overlapping raw significant tiles
      wlog(debugLevel, 2, "  selecting raw events");
      rawEvents.wselect(rawSignificants, pchan.plotDurationInflation, 
			pchan.plotBandwidthInflation, pchan.maximumMosaics, 
			debugLevel);
   }

   //////////////////////////////////////////////////////////////////////////
   //                     add channel to summary report                    //
   //////////////////////////////////////////////////////////////////////////

   // identify most significant raw transform tile
   wlog(debugLevel, 2, "  measuring peak properties");
   thresholdReferenceTime = whitenedProperties[0].peak.centerTime();
   thresholdTimeRange[0] = -whitenedProperties[0].peak.Duration;
   thresholdTimeRange[1] = -thresholdTimeRange[0];
   thresholdFrequencyRange.resize(2);
   thresholdFrequencyRange[0] = whitenedProperties[0].peak.Frequency 
      - whitenedProperties[0].peak.Bandwidth;
   thresholdFrequencyRange[1] = whitenedProperties[0].peak.Frequency 
      + whitenedProperties[0].peak.Bandwidth;
   thresholdQRange = emptyRange;
   wmeasure rawProperties(rawTransform, tiling, startTime, 
			  thresholdReferenceTime, thresholdTimeRange, 
			  thresholdFrequencyRange, thresholdQRange, 
			  debugLevel);
   //rawProperties.dump(cout);
 
   // most significant tile properties
   Time mostSignificantTime =  whitenedProperties[0].peak.centerTime();
   double mostSignificantFrequency = whitenedProperties[0].peak.Frequency;
   double mostSignificantQ =  whitenedProperties[0].peak.Q;
   double mostSignificantNormalizedEnergy =
      whitenedProperties[0].peak.NormalizedEnergy;

   //----------------------------------------------------------------------
   //   This was changed to use signalAmplitude[0]
   // double mostSignificantAmplitude = rawProperties{1}.peakAmplitude;
   // double mostSignificantAmplitude = rawProperties[0].signalAmplitude;
   double mostSignificantAmplitude = rawProperties[0].signal.Amplitude;

   // write most significant tile properties to text summary file
   wlog(debugLevel, 2, "  writing text summary");
   cerr << "Channel name:                        " << channelNameString << endl;
   cerr << "Most significant tile time:          " 
	<< mostSignificantTime.totalS() << endl;
   cerr << "Most significant tile Frequency:     " << mostSignificantFrequency 
	<< endl;
   cerr << "Most significant tile Q:             " << mostSignificantQ 
	<< endl; 
   cerr << "Most significant  tile normalized E: " 
	<< mostSignificantNormalizedEnergy << endl;
   cerr << "Most significant tile amplitude: " << mostSignificantAmplitude 
	<< endl;

   //===================================  begin loop over time ranges
   for (size_t iPlTiR=0; iPlTiR < plotTimeRanges.size(); iPlTiR++) {
      double plotTimeRange =  plotTimeRanges[iPlTiR];
      dble_vect autoNormalizedEnergyRange;

      // report status
      if (debugLevel >= 2) {
	 cout << "  processing " << plotTimeRange << " second time range" 
	      << endl;
      }

      dble_vect tRange(2);
      tRange[0] = -plotTimeRange/2;
      tRange[1] =  plotTimeRange/2;

      dble_vect qRange(1, mostSignificantQ);
	 
      //--------------------------  Loop over plots;
      int nPlot = pchan.plotType.size();
      for (int iplot=0; iplot< nPlot; iplot++) {
	 wfigure figureHandle;
	 figureHandle.open();
	 string pltType = pchan.plotType[iplot];
	 if (pltType == tser_raw) {
	    figureHandle.wtimeseries(rawData[0], 
				     centerTime, 
				     -plotTimeRange/2,
				     plotTimeRange/2, 
				     channelNameString);
	 }
	 else if (pltType == tser_hipas) {
	    figureHandle.wtimeseries(conditioner.highPassedData(), 
				     centerTime, 
				     -plotTimeRange/2, plotTimeRange/2, 
				     channelNameString);
	 }
	 else if (pltType == tser_white) {
	    figureHandle.wtimeseries(conditioner.whitenedData(), 
				     centerTime, 
				     -plotTimeRange/2, plotTimeRange/2, 
				     channelNameString);
	 }
	 else if (pltType == spec_raw) {
	    figureHandle.wspectrogram(rawTransform[0], tiling, 
				      centerTime, 
				      tRange, plotFrequencyRange, 
				      qRange, plotNormalizedEnergyRange, 
				      pchan.plotHorizontalResolution);
	 }
	 else if (pltType == spec_white) {
	    figureHandle.wspectrogram(whitenedTransform[0], tiling,
				      centerTime, tRange, 
				      plotFrequencyRange, 
				      qRange, plotNormalizedEnergyRange, 
				      pchan.plotHorizontalResolution);
	 }
	 else if (pltType == spec_auto) {
	    figureHandle.wspectrogram(whitenedTransform[0], tiling, 
				      centerTime, tRange,
				      plotFrequencyRange, qRange, 
				      autoNormalizedEnergyRange, 
				      pchan.plotHorizontalResolution);
	 }
	 else if (pltType == evnt_raw) {
	    figureHandle.weventgram(rawEvents[0], tiling, centerTime, 
				    tRange, plotFrequencyRange, 
				    pchan.plotDurationInflation,
				    pchan.plotBandwidthInflation, 
				    plotNormalizedEnergyRange);
	 }
	 else if (pltType == evnt_white) {
	    figureHandle.weventgram(whitenedEvents[0], tiling, centerTime,
				    tRange, plotFrequencyRange, 
				    pchan.plotDurationInflation, 
				    pchan.plotBandwidthInflation, 
				    plotNormalizedEnergyRange);
	 }
	 else if (pltType == evnt_auto) {
	    figureHandle.weventgram(whitenedEvents[0], tiling, centerTime, 
				    tRange, plotFrequencyRange, 
				    pchan.plotDurationInflation, 
				    pchan.plotBandwidthInflation, 
				    autoNormalizedEnergyRange);
	 }
	 else {
	    cerr << "Unrecognized plot type: " << pltType << endl;
	    continue;
	 }
	 ostringstream figName;
	 if (pchan.plotFile.empty()) {
	    figName << eventTimeString << "_" << channelNameString << "_" 
		    << fixed << setprecision(2) << plotTimeRange 
		    << "_" << pltType;
	    outFile.set_prefix(figName.str());
	 } else if (nPlot > 1) {
	    figName << pchan.plotFile << "_" << pltType;
	    outFile.set_prefix(figName.str());
	 }
	 figureHandle.wprintfig(outFile.file_path(pchan.eventTime), 
				pchan.plotThumbnail);
      }
   }   // end loop time ranges


   //-----------------------------------  Exit
   if (debugLevel >= 1) {
      cout << "finished on " << datestr(0, 29) 
	   << " at " << datestr(0, 13) << endl;
   }
   //figureHandle.close();
}

//======================================  Search channel constructor.
plot_chan::plot_chan(void) {
   clear();
}

//======================================  Plot type list
static const char* plot_types[] = {
      "spectrogram_whitened", "spectrogram_raw", "spectrogram_autoscaled",
      "timeseries_raw", "timeseries_whitened", "timeseries_hipassed", 
      "eventgram_whitened", "eventgram_raw", "eventgram_autoscaled"};


//======================================  Clear search channel descriptor.
void
plot_chan::clear(void) {
   channelName.clear();
   eventTime = Time(0);
   frameType.clear();
   sampleFrequency = 0;
   searchTimeRange = 0;
   searchFrequencyRange.clear();
   searchQRange.clear();
   searchMaximumEnergyLoss = 0;
   whiteNoiseFalseRate = 0;
   eventThreshold = 0;
   searchWindowDuration = 0;
   plotTimeRanges.clear();
   plotFrequencyRange.clear();
   plotNormalizedEnergyRange.clear();
}

//======================================  Set configuration struct to defaults
void
plot_chan::reset(void) {
   clear();
   const double dInf(1.0/0.0);
   channelName.clear();
   frameType = "NDS2";
   eventTime = Time(0);
   sampleFrequency = 4096;
   searchTimeRange = 64;
   double fRange[] = {0, dInf};
   searchFrequencyRange = dble_vect(fRange, fRange+2);
   double qRange[] = {4, 64};
   searchQRange = dble_vect(qRange, qRange+2);
   searchMaximumEnergyLoss = 0.2;
   whiteNoiseFalseRate = 1;
   searchWindowDuration = 0.1;
   double plotRange[] = {1, 4, 16};
   plotTimeRanges = dble_vect(plotRange, plotRange+1);
   double plotFRange[]  = {0, dInf};
   plotFrequencyRange = dble_vect(plotFRange, plotFRange+2);
   //notused double defaultPlotMaximumEnergyLoss = 0.2;
   double plotNormER[] = {0, 25.5};
   plotNormalizedEnergyRange = dble_vect(plotNormER, plotNormER+2);

   // plot types
   plotType = str_vect(plot_types, plot_types+1);
   plotThreshold = 0.0;

   // search parameters
   transientFactor = 2.0;
   outlierFactor = 2.0;

   // display parameters
   plotHorizontalResolution = 512;
   plotDurationInflation = 0.5;
   plotBandwidthInflation = 0.5;

   // limits on number of significant tiles
   maximumSignificants = 10000;
   maximumMosaics = 10000;

   plotThumbnail = false;
}

//======================================  Scan the configuration.
void
wpipe::scan_config(int argc, const char* argv[], plot_chan& p) {

   for (int iarg=1; iarg<argc; iarg++) {
      string cmdstr = argv[iarg];
      if (cmdstr == "--help") {
	 wplot_syntax();
	 error("Help requested");
      }

      // locate field separator
      string::size_type colonIndex = cmdstr.find_first_of(":=");

      // if (field separator not located, report syntax error
      if (colonIndex == string::npos) {
	 cerr << "syntax error parsing command line: " << cmdstr << endl;
	 error("syntax error processing configuration file");
      }
    
      // parse configuration line
      string parameterName = cmdstr.substr(0, colonIndex);
      string parameterValue = cmdstr.substr(colonIndex + 1);
      const char* parValue = parameterValue.c_str();

      //------------------------------------  Get channel name
      if (parameterName == "channelName") {
	 p.channelName = unquote(parameterValue);
      } 

      //------------------------------------  Get the start time.
      else if (parameterName == "eventTime") {
	 double time = strtod(parValue, 0);
	 long sec = long(time);
	 p.eventTime = Time(sec, long((time-sec)*1e9));
      }

      //------------------------------------  Run parameters
      else if (parameterName == "debugLevel") {
	 p.debugLevel = strtol(parValue, 0, 0);
      } 
      else if (parameterName == "plotThumbnail") {
	 p.plotThumbnail = (strtol(parValue, 0, 0) != 0);
      } 
      else if (parameterName == "outputDirectory") {
	 p.outDir = parValue;
      }
      else if (parameterName == "plotFile") {
	 p.plotFile = parValue;
      }

      //------------------------------------  Data location parameters
      else if (parameterName == "frameType") {
	 p.frameType = unquote(parameterValue);
      } else if (parameterName == "ndsServer") {
	 p.ndsServer = unquote(parameterValue);
      } else if (parameterName ==  "sampleFrequency") {
	 p.sampleFrequency = strtod(parValue, 0);
      } else if (parameterName ==  "searchTimeRange") {
	 p.searchTimeRange = strtod(parValue, 0);
      } else if (parameterName ==  "searchFrequencyRange") {
	 p.searchFrequencyRange = eval(parameterValue);
      } else if (parameterName ==  "searchQRange") {
	 p.searchQRange = eval(parameterValue);
      } else if (parameterName ==  "searchMaximumEnergyLoss") {
	 p.searchMaximumEnergyLoss = strtod(parValue, 0);
      } else if (parameterName ==  "whiteNoiseFalseRate") {
	 p.whiteNoiseFalseRate = strtod(parValue, 0);
      } else if (parameterName ==  "eventThreshold") {
	 p.eventThreshold = strtod(parValue, 0);
      } else if (parameterName ==  "plotThreshold") {
	 p.plotThreshold = strtod(parValue, 0);
      } else if (parameterName ==  "searchWindowDuration") {
	 p.searchWindowDuration = strtod(parValue, 0);
      } else if (parameterName ==  "plotTimeRanges") {
	 p.plotTimeRanges = eval(parameterValue);
      } else if (parameterName ==  "plotFrequencyRange") {
	 p.plotFrequencyRange = eval(parameterValue);
      } else if (parameterName ==  "plotNormalizedEnergyRange") {
	 p.plotNormalizedEnergyRange = eval(parameterValue);
      } else if (parameterName == "plotType") {
	 p.plotType = strsplit(deblank(parameterValue), ",");
	 for (const_strv_iter i=p.plotType.begin(); i!= p.plotType.end(); i++) {
	    bool ok = false;
	    for (const char** j=plot_types; *j; j++) {
	       if (*i == *j) {
		  ok = true;
		  break;
	       }
	    }
	    if (!ok) error(string("Invalid plot type name: ")+*i);
	 }
      }

      //--------------------------------  Invalid parameter name
      else {
	 cerr << "unknown configuration parameter " << parameterName << endl;
	 error("unknown configuration parameter");
      }

      // end loop over channel configuration file
   }
}

//======================================  Diaplay a numeric vector parameter.
static void
display_dble_vect(const dble_vect& dv, std::ostream& out) {
   size_t N = dv.size();
   out << "[";
   for (size_t i=0; i<N; i++) {
      if (i) out << " ";
      out << dv[i];
   }
   out << "]" << endl;
}

//======================================  Diaplay channel parameters.
void
plot_chan::display(std::ostream& out) const {
   out << "channelName:               " << channelName << endl;
   out << "frameType:                 " << frameType << endl;
   out << "sampleFrequency:           " << sampleFrequency << endl;
   out << "searchTimeRange:           " << searchTimeRange << endl;
   out << "searchFrequencyRange:      "; 
   display_dble_vect(searchFrequencyRange, out);
   out << "searchQRange:              "; display_dble_vect(searchQRange, out);
   out << "searchMaximumEnergyLoss:   " << searchMaximumEnergyLoss << endl;
   out << "whiteNoiseFalseRate:       " << whiteNoiseFalseRate << endl;
   out << "searchWindowDuration:      " << searchWindowDuration << endl;
   out << "plotTimeRange:             "; display_dble_vect(plotTimeRanges, out);
   out << "plotFrequencyRange:        "; 
   display_dble_vect(plotFrequencyRange, out);
   out << "plotNormalizedEnergyRange: ";
   display_dble_vect(plotNormalizedEnergyRange, out);
}
